@@ -70,23 +70,20 @@
 #
 # See: https://github.com/python-trio/trio/issues/53
 
-from itertools import count
 import attr
 from collections import OrderedDict
 
 from .. import _core
-from .._util import SubclassingDeprecatedIn_v0_15_0
-
-_counter = count()
+from .._util import Final
 
 
-@attr.s(frozen=True)
+@attr.s(frozen=True, slots=True)
 class _ParkingLotStatistics:
     tasks_waiting = attr.ib()
 
 
-@attr.s(eq=False, hash=False)
-class ParkingLot(metaclass=SubclassingDeprecatedIn_v0_15_0):
+@attr.s(eq=False, hash=False, slots=True)
+class ParkingLot(metaclass=Final):
     """A fair wait queue with cancellation and requeueing.
 
     This class encapsulates the tricky parts of implementing a wait
